from slobot.feetech_frame import FeetechFrame
from slobot.configuration import Configuration

import rerun as rr

class Metrics():
    LOGGER = Configuration.logger(__name__)

    def __init__(self):
        rr.init("teleoperation", spawn=True)
        rr.log("qpos", rr.SeriesLines(names=self.metric_names("Follower")), static=True)
        rr.log("target_qpos", rr.SeriesLines(names=self.metric_names("Leader")), static=True)
        rr.log("velocity", rr.SeriesLines(names=self.metric_names("Velocity")), static=True)
        rr.log("force", rr.SeriesLines(names=self.metric_names("Force")), static=True)
        self.step = 0

    def metric_names(self, metric_name):
        return [
            f"{metric_name} {joint_name}"
            for joint_name in Configuration.JOINT_NAMES
        ]

    def handle_qpos(self, feetech_frame: FeetechFrame):
        Metrics.LOGGER.debug(f"Feetech frame {feetech_frame}")

        rr.set_time("step", sequence=self.step)
        rr.log("qpos", rr.Scalars(feetech_frame.qpos))
        rr.log("target_qpos", rr.Scalars(feetech_frame.target_qpos))
        rr.log("velocity", rr.Scalars(feetech_frame.velocity))
        rr.log("force", rr.Scalars(feetech_frame.force))

        self.step += 1



import numpy as np
import logging
from importlib.resources import files

class Configuration:
    WORK_DIR = "/tmp/slobot"

    MJCF_CONFIG = str(files('slobot.config') / "trs_so_arm100" / "so_arm100.xml")

    # 16:9 aspect ratio
    LD = (426, 240)
    SD = (854, 480)
    HD = (1280, 720)
    FHD = (1920, 1080)

    QPOS_MAP = {
        "middle": [0, -np.pi/2, np.pi/2, 0, 0, -0.15],
        "zero": [0, 0, 0, 0, 0, 0],
        "rotated": [-np.pi/2, -np.pi/2, np.pi/2, np.pi/2, -np.pi/2, np.pi/2],
        "rest": [0.049, -3.32, 3.14, 1.21, -0.17, -0.17]
    }

    POS_MAP = {
        "middle": [2095, 2095, 2095, 2095, 2095, 2095],
        "zero": [2097, 3143, 978, 1997, 1081, 2162],
        "rotated": [3161, 2126, 2016, 3017, 12, 3330],
        "rest": [2089, 965, 3063, 2822, 1035, 2030]
    }

    DOFS = 6
    JOINT_NAMES = ["shoulder_pan", "shoulder_lift", "elbow_flex", "wrist_flex", "wrist_roll", "gripper"]

    def logger(logger_name):
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logger.addHandler(handler)
        return logger


from lerobot.common.datasets.v2.convert_dataset_v1_to_v2 import make_robot_config
from lerobot.common.motors.feetech import TorqueMode
from lerobot.common.robots import make_robot_from_config
from lerobot.common.motors import MotorsBus

from slobot.configuration import Configuration
from slobot.simulation_frame import SimulationFrame
from slobot.feetech_frame import FeetechFrame

import json
import numpy as np
import time

class Feetech():
    ROBOT_TYPE = 'so100_follower'
    FOLLOWER_ID = 'follower_arm'
    LEADER_ID = 'leader_arm'

    MOTOR_MODEL = 'sts3215'
    MOTOR_DIRECTION = [-1, 1, 1, 1, 1, 1]
    JOINT_IDS = range(Configuration.DOFS)

    PORT0 = '/dev/ttyACM0'
    PORT1 = '/dev/ttyACM1'

    REFERENCE_FRAME = 'middle'

    def calibrate_pos(preset):
        feetech = Feetech()
        feetech.calibrate(preset)

    def move_to_pos(pos):
        feetech = Feetech()
        feetech.control_position(pos)

    def __init__(self, **kwargs):
        self.port = kwargs.get('port', Feetech.PORT0)
        self.robot_id = kwargs.get('robot_id', Feetech.FOLLOWER_ID)
        self.qpos_handler = kwargs.get('qpos_handler', None)
        connect = kwargs.get('connect', True)
        torque = kwargs.get('torque', True)

        self.motors_bus : MotorsBus = self._create_motors_bus(self.port, self.robot_id)
        if connect:
            self.connect(torque)

    def connect(self, torque):
        self.motors_bus.connect()
        if torque:
            self.set_torque(True)

    def disconnect(self):
        self.set_torque(False)
        self.motors_bus.disconnect()

    def get_qpos(self):
        return self.pos_to_qpos(self.get_pos())

    def get_pos(self, ids = JOINT_IDS):
        return self._read_config('Present_Position', ids)

    def get_velocity(self):
        return self._read_config('Present_Velocity')

    def get_dofs_velocity(self):
        return self.velocity_to_qvelocity(self.get_velocity())

    def get_dofs_control_force(self, ids = JOINT_IDS):
        return self._read_config('Present_Load', ids=ids)
    
    def get_pos_goal(self):
        return self._read_config('Goal_Position')

    def handle_step(self, frame: SimulationFrame):
        pos = self.qpos_to_pos(frame.qpos)
        self.control_position(pos)

    def qpos_to_pos(self, qpos):
        return [ self._qpos_to_steps(qpos, i)
            for i in Feetech.JOINT_IDS ]

    def pos_to_qpos(self, pos, ids = JOINT_IDS):
        return [ self._steps_to_qpos(pos, id)
            for id in ids]

    def velocity_to_qvelocity(self, velocity):
        return [ self._stepvelocity_to_velocity(velocity, i)
            for i in Feetech.JOINT_IDS ]

    def control_position(self, pos, ids=JOINT_IDS):
        self._write_config('Goal_Position', pos, ids)
        if self.qpos_handler is not None:
            feetech_frame = self.create_feetech_frame(pos)
            self.qpos_handler.handle_qpos(feetech_frame)

    def control_dofs_position(self, target_qpos):
        target_pos = self.qpos_to_pos(target_qpos)
        self.control_position(target_pos)

    def get_torque(self, ids=JOINT_IDS):
        return self._read_config('Torque_Enable', ids)

    def set_torque(self, is_enabled: bool, ids=JOINT_IDS):
        torque_enable = TorqueMode.ENABLED.value if is_enabled else TorqueMode.DISABLED.value
        torque_enable = [
            torque_enable
            for joint_id in ids
        ]
        self._write_config('Torque_Enable', torque_enable, ids)

    def set_home_offset(self, home_offset, ids=JOINT_IDS):
        self._write_config("Home_Offset", home_offset, ids)

    def set_punch(self, punch, ids=JOINT_IDS):
        self._write_config('Minimum_Startup_Force', punch, ids)

    def set_dofs_kp(self, Kp, ids=JOINT_IDS):
        self._write_config('P_Coefficient', Kp, ids)

    def get_dofs_kp(self, ids=JOINT_IDS):
        return self._read_config('P_Coefficient', ids)

    def set_dofs_kv(self, Kv, ids=JOINT_IDS):
        self._write_config('D_Coefficient', Kv, ids)

    def get_dofs_kv(self, ids=JOINT_IDS):
        return self._read_config('D_Coefficient', ids)

    def set_dofs_ki(self, Ki, ids=JOINT_IDS):
        self._write_config('I_Coefficient', Ki, ids)

    def get_dofs_ki(self, ids=JOINT_IDS):
        return self._read_config('I_Coefficient', ids)

    def go_to_rest(self):
        self.go_to_preset('rest')

    def go_to_preset(self, preset):
        pos = Configuration.POS_MAP[preset]
        self.move(pos)
        time.sleep(1)
        self.disconnect()

    def calibrate(self, preset):
        self.set_torque(False)
        input(f"Move the arm to the {preset} position ...")
        pos = self.get_pos()
        pos_json = json.dumps(pos)
        print(f"Current position is {pos_json}")

    def _create_motors_bus(self, port, robot_id) -> MotorsBus:
        robot_config = make_robot_config(Feetech.ROBOT_TYPE, port=port, id=robot_id)
        robot = make_robot_from_config(robot_config)
        motors_bus = robot.bus

        self.model_resolution = motors_bus.model_resolution_table[Feetech.MOTOR_MODEL]
        self.radian_per_step = (2 * np.pi) / self.model_resolution

        return motors_bus

    def _qpos_to_steps(self, qpos, motor_index):
        steps = Feetech.MOTOR_DIRECTION[motor_index] * (qpos[motor_index] - Configuration.QPOS_MAP[Feetech.REFERENCE_FRAME][motor_index]) / self.radian_per_step
        return Configuration.POS_MAP[Feetech.REFERENCE_FRAME][motor_index] + int(steps)

    def _steps_to_qpos(self, pos, motor_index):
        steps = pos[motor_index] - Configuration.POS_MAP[Feetech.REFERENCE_FRAME][motor_index]
        return Configuration.QPOS_MAP[Feetech.REFERENCE_FRAME][motor_index] + Feetech.MOTOR_DIRECTION[motor_index] * steps * self.radian_per_step

    def _stepvelocity_to_velocity(self, step_velocity, motor_index):
        return step_velocity[motor_index] * self.radian_per_step

    def _read_config(self, key, ids=JOINT_IDS):
        motors = [
            Configuration.JOINT_NAMES[id]
            for id in ids
        ]
        pos = self.motors_bus.sync_read(key, motors, normalize=False)
        return [
            pos[Configuration.JOINT_NAMES[id]]
            for id in ids
        ]

    def _write_config(self, key, values, ids):
        values = {
            Configuration.JOINT_NAMES[id] : values[i]
            for i, id in enumerate(ids)
        }
        self.motors_bus.sync_write(key, values, normalize=False)

    def create_feetech_frame(self, target_pos) -> FeetechFrame:
        timestamp = time.time()
        qpos = self.pos_to_qpos(self.get_pos())
        target_qpos = self.pos_to_qpos(target_pos)
        velocity = self.get_dofs_velocity()
        force = self.get_dofs_control_force()
        return FeetechFrame(timestamp, target_qpos, qpos, velocity, force)


import time

import torch
import cv2
import numpy as np

from lerobot.common.datasets.v2.convert_dataset_v1_to_v2 import make_robot_config
from lerobot.common.policies.factory import get_policy_class
from lerobot.common.robots import make_robot_from_config
from lerobot.common.utils.robot_utils import busy_wait
from lerobot.common.utils.utils import auto_select_torch_device


class PolicyEvaluator:
    INFERENCE_TIME_S = 60
    FPS = 25

    def __init__(self, robot_type, policy_type, model_path, port):
        self.model_path = model_path
        self.policy_type = policy_type

        robot_config = make_robot_config(robot_type, port=port)
        self.robot = make_robot_from_config(robot_config)

        self.device = auto_select_torch_device()

    def evaluate(self):
        self.robot.connect()

        policy_cls = get_policy_class(self.policy_type)

        self.policy = policy_cls.from_pretrained(self.model_path)
        self.policy.to(self.device)

        for step in range(self.INFERENCE_TIME_S * self.FPS):
            self.run_step(step)

    def run_step(self, step):
        start_time = time.perf_counter()

        # Read the follower state and access the frames from the cameras
        observation = self.robot.capture_observation()

        image = observation['observation.images.phone']
        np_image = np.array(image)
        np_image = cv2.cvtColor(np_image, cv2.COLOR_RGB2BGR)
        cv2.imwrite(f"eval_images/img_{step:04d}.jpg", np_image)

        # Convert to pytorch format: channel first and float32 in [0,1]
        # with batch dimension
        for name in observation:
            if "image" in name:
                observation[name] = observation[name].type(torch.float32) / 255
                observation[name] = observation[name].permute(2, 0, 1).contiguous()
            observation[name] = observation[name].unsqueeze(0)
            observation[name] = observation[name].to(self.device)

        # Compute the next action with the policy
        # based on the current observation
        action = self.policy.select_action(observation)
        # Remove batch dimension
        action = action.squeeze(0)
        # Move to cpu, if not already the case
        action = action.to("cpu")
        # Order the robot to move
        self.robot.send_action(action)

        dt_s = time.perf_counter() - start_time
        busy_wait(1 / self.FPS - dt_s)